--Functions_non_spec.lua contains the functions that are not specific to envelope type
--i.e. vol pan dly
require "Zoom"
----------------
--FUNCTION LIST
----------------
--create_automation_envelopes_all_tracks() 
--update_selected_envelope_from_pattern()
--alt_undo(undo)
--set_flag(bool) 
--check_for_present_dummy_devices()
--add_notifiers_and_timers()
--remove_dummy_devices()

-------------
--CONSTANTS 
------------
local EMPTY_NOTE = renoise.PatternTrackLine.EMPTY_NOTE --121
local NOTE_OFF = renoise.PatternTrackLine.NOTE_OFF --120

--value constants
VEL_MAX = 128
VEL_EMPTY = 255

PAN_MAX = 128
PAN_CENTRE = 64 
PAN_EMPTY = 255

DLY_MAX = 255
DLY_EMPTY = 0

----------
--GlOBALS
----------
flag = false --global flag used for timer notifier --if true updating a pattern envelope will update pattern
flag_all_columns = false --all columns with corresponding notes
--globals used in timer notifier for:  check_selected_column_changed()
g_selected_column = nil
g_selected_track = nil
g_selected_pattern = nil
g_selected_parameter_name = ""

--GUI Global
my_dialog = nil --don`t declare as local as not in main file 
vb = renoise.ViewBuilder()

------------
--FUNCTIONS
------------
--------------------------------------------
--------[[function sets up tool]]-----------========================================================================
--------------------------------------------
--adds dummy devices to all tracks
--and sets up envelopes for use
--called when GUI is first open
--------------------------------------------
--------------------------------------------
function create_automation_envelopes_all_tracks() 
  
  --reset global flag all columns to false in case needed
  flag_all_columns = false
  flag = false
    
  --[2] set selected parameter not possible!!! not in API
   
  --[3] add notifiers
  --------------------------------------------------------------
  --check there is a timer for selected column, if not add it.  This checks if
  --the selected column/track/pattern has changed and refreshes the automation
  --view to the new columns values if it has. 
  --(mimics a song.selected_column_observable notifier) (also pattern/track)
  --------------------------------------------------------------
  --check_column_has_changed() calls update_selected_envelope_from_pattern() 
  --(pattern -> envelope)
 
  if not renoise.tool():has_timer(check_selected_column_changed) then
   renoise.tool():add_timer(check_selected_column_changed,100) 
  end
 
  --------------------------------------------------------------------------
  --add timer to trigger write_envelope_to_pattern() every X milliseconds
  --if global variable `flag` is set to false nothing will happen
  --if set to true the function will copy the automation data to the pattern
  --------------------------------------------------------------------------
  -- write_envelope_to_pattern() does the opposite and writes envelope to pattern
 
 
  if not renoise.tool():has_timer(write_envelope_to_pattern) then 
    renoise.tool():add_timer(write_envelope_to_pattern,300)         --- --THIS CAUSES BUGGY BEHAVIOUR!!!!!! <-> etc.
  end                                    
  -----------------------------------------------------------
 
  --renoise.song()
  local song = renoise.song()
  
  --gives the parameter number in relation to the device it is in. 
  --this is not the automation number.  Automation number is a dynamic table (stack?) that contains the latest user added automations. 
  local sel = song.selected_parameter
  --selected indexes
  local pattern = song.selected_pattern_index
  local track = song.selected_track_index
  local column = song.selected_note_column_index
  local device = renoise.song().selected_device_index
  local selected_param = renoise.song().selected_automation_parameter--song.selected_parameter_index

  --selected pattern_track + its length
  local selected_pattern_track = song.patterns[pattern].tracks[track]
  local target_pattern_track = selected_pattern_track
  local sel_pattern_lines = renoise.song().patterns[pattern].number_of_lines
  
  --reset timer notifier variables
  g_selected_column = column
  g_selected_track = track
  g_selected_pattern = pattern
 
  --parameter name can only be set if selected else set it to empty string
  if song.selected_parameter then 
    g_selected_parameter_name = song.selected_parameter.name
  else
    g_selected_parameter_name = ""
  end
    
  --Loop through all the tracks and add a dummy "Vel Pan Dly" device to create automation envelopes from
  --the device is a *Formula device which serves as a placeholder.  (The *formula device is used as it can have it`s parameter
  --names changed to "Vel" "Pan" and "Dly" and has no direct effect on audio)
  
  for trk = 1, #song.tracks do
    --make sure track is a sequencer track
    if song.tracks[trk].type > 1 then
     --do nothing for this iteration of the loop as not a sequencer track
    else 
      --set target pattern track
      target_pattern_track = song.patterns[pattern].tracks[trk]
    
      --loop to check if "Vel Pan Dly" DSP has already been added
      local found_device = false
      for i = 2,#song.tracks[trk].devices do
        if song.tracks[trk].devices[i].display_name == "Vel Pan Dly" then
          found_device = true
          break
        end
      end
      
      --create variable to hold the dummy devie
      local dummy_device 
      --if not found, add "Vel Pan Dly" DSP -- which is a dummy stereo expander --MAY CHANGE
      if found_device == false then
       
        -- set lower frame to view automation
        renoise.app().window.active_lower_frame = renoise.ApplicationWindow.LOWER_FRAME_TRACK_AUTOMATION
        
         --insert dummy device
         dummy_device = song.tracks[trk]:insert_device_at("Audio/Effects/Native/*Formula", 2)
         dummy_device.display_name = "Vel Pan Dly"
          --set the values for the device including Velocity Panning And Delay as perameter names
         dummy_device.active_preset_data =
              
            [[<?xml version="1.0" encoding="UTF-8"?>
          <FilterDevicePreset doc_version="9">
            <DeviceSlot type="FormulaMetaDevice">
              <IsMaximized>false</IsMaximized>
              <FormulaParagraphs>
                <FormulaParagraph>A * sin(LINEF / NUMLINES * TWOPI * B) + C</FormulaParagraph>
              </FormulaParagraphs>
              <FunctionsParagraphs>
                <FunctionsParagraph>function inv(x)</FunctionsParagraph>
                <FunctionsParagraph>  return (1.0 - x)</FunctionsParagraph>
                <FunctionsParagraph>end</FunctionsParagraph>
                <FunctionsParagraph/>
              </FunctionsParagraphs>
              <InputNameA>VEL</InputNameA>
              <InputNameB>PAN</InputNameB>
              <InputNameC>DLY</InputNameC>
              <EditorVisible>false</EditorVisible>
              <PanelVisible>0</PanelVisible>
              <InputA>
                <Value>0.0</Value>
              </InputA>
              <InputB>
                <Value>0.0</Value>
              </InputB>
              <InputC>
                <Value>0.0</Value>
              </InputC>
            </DeviceSlot>
          </FilterDevicePreset>]]
    
        --create an automation for VEL PAN DLY
        target_pattern_track:create_automation(dummy_device.parameters[1]) --VEL
        target_pattern_track:create_automation(dummy_device.parameters[2]) --PAN
        target_pattern_track:create_automation(dummy_device.parameters[3]) --DLY
      
        --loop through automations (stack?) to get the automation number for the "Velocity Device" - Input parameter "VEL"
        local target_automation = nil
        for i = 1, #target_pattern_track.automation do
          if target_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
            if target_pattern_track.automation[i].dest_parameter.name == "VEL" then
              target_automation = i
            end
          end
        end
        
        --set playmode to points for VEL,PAN,DLY
        target_pattern_track.automation[target_automation].playmode = renoise.PatternTrackAutomation.PLAYMODE_POINTS --vel
        target_pattern_track.automation[target_automation+1].playmode = renoise.PatternTrackAutomation.PLAYMODE_POINTS --pan
        target_pattern_track.automation[target_automation+2].playmode = renoise.PatternTrackAutomation.PLAYMODE_POINTS --dly
                                    
      end --for loop 
    end--if device found 
  end--if sequencer track 
end-- create_automation_envelopes_all_tracks() function

-----------------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------

----------------------------------------------------------
--gets called by selected parameter(== envelope) notifier
--TODO:
--removes previous notifiers
--checks if envelope is vol,pan,dly --return if not
--adds relevant notifiers
----------------------------------------------------------

function update_selected_envelope_from_pattern()
  
  local song = renoise.song()
    
  --check if device is vel,pan,dly automated
  --loop through automations (stack?) to get the automation number for the "Velocity Device" - Input parameter
  
  --set target pattern track
  local target_pattern_track = song.selected_pattern_track
  local target_automation = nil
  for i = 1, #target_pattern_track.automation do
    if target_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
      if target_pattern_track.automation[i].dest_parameter.name == "VEL" then
        target_automation = i
      end
    end
  end
  
  --if selected column not a note column clear all possible envelopes then return
  if song.selected_sub_column_type >= 6 then --6 and 7 is are fx column sub type ---CAN BE MOVED 
    song.selected_pattern_track.automation[target_automation]:clear()
    song.selected_pattern_track.automation[target_automation+1]:clear()
    song.selected_pattern_track.automation[target_automation+2]:clear()
    return
  end
  
  --check if selected envelope is vol,pan,dly, if not return
  local p_name
  --check if parameter selected if not return
  if song.selected_parameter then
    p_name = song.selected_parameter.name
  else
  renoise.app():show_status("`Vel Pan Dly` Tool: (No Parameter Selected, please select one)") 
    return
  end
  
  --return if not at least one of the following ((note double bracket)) envelope names is the selected parameter(==envelope)
  if not ((p_name == "VEL") or (p_name == "PAN") or (p_name == "DLY")) then
    return
  end
  
  --current pattern length for loop max
  local sel_pattern_lines = song.selected_pattern.number_of_lines
  local column = song.selected_note_column_index
  
  --remove notifiers
  remove_notifiers_and_timers()
  
  --p_name now gives us the envelope type.
  if p_name == "VEL" then
    for line = 1, sel_pattern_lines do
      --get velocity value from pattern and divide by their range (envelope values are between 0-1)
      local line_vel = target_pattern_track:line(line).note_columns[column].volume_value/VEL_MAX
      --if the velocity column is empty ("--") then set line_vol to 1 (envelope maximum)
      if line_vel == (VEL_EMPTY/VEL_MAX) then
        line_vel = 1
      end
       
      --remove points if no note is present
      if target_pattern_track:line(line).note_columns[column].note_string == "OFF" or
        target_pattern_track:line(line).note_columns[column].note_string == "---" then 
        if target_pattern_track.automation[target_automation]:has_point_at(line) then
          target_pattern_track.automation[target_automation]:remove_point_at(line)
        end
      else
        --write the pattern velocity to the envelope value
        if line_vel <= 1 and line_vel >= 0 then
           --scale envelope if preference chosen by user
           line_vel = wte_zoom_env_lower_half(line_vel)
           target_pattern_track.automation[target_automation]:add_point_at(line,line_vel) 
        end
      end
    end
  end--"VEL"----
  
  if p_name == "PAN" then 
    for line = 1, sel_pattern_lines do
      --get pan value from pattern and divide by their range (envelope values are between 0-1)
      local line_pan = target_pattern_track:line(line).note_columns[column].panning_value/PAN_MAX --renoise has double the range for pan values
      --if the pan column is empty ("--") then set line_pan to 0.5 (envelope mid-point)
      if line_pan == (PAN_EMPTY/PAN_MAX) then
        line_pan = 0.5
      end
      --remove points if no note is present
      if target_pattern_track:line(line).note_columns[column].note_string == "OFF" or
        target_pattern_track:line(line).note_columns[column].note_string == "---" then 
        if target_pattern_track.automation[target_automation+1]:has_point_at(line) then
          target_pattern_track.automation[target_automation+1]:remove_point_at(line)
        end
      else
      --pan
        if line_pan <= 1 and line_pan >= 0 then
          --scale envelope if preference chosen by user
          line_pan = wte_zoom_env_lower_half(line_pan)
          target_pattern_track.automation[target_automation+1]:add_point_at(line,line_pan) 
        end
      end
    end
  end --"PAN"----
  
  if p_name == "DLY" then
    for line = 1, sel_pattern_lines do
      --get delay value from pattern and divide by their range (envelope values are between 0-1)
      local line_dly = target_pattern_track:line(line).note_columns[column].delay_value/DLY_MAX --and delay values
      --if the delay column is empty ("--") then set line_del to 0 (envelope zero)
      if line_dly == (DLY_EMPTY/DLY_MAX) then
        line_dly = 0
      end
      --remove points if no note is present
      if target_pattern_track:line(line).note_columns[column].note_string == "OFF" or
        target_pattern_track:line(line).note_columns[column].note_string == "---" then 
        if target_pattern_track.automation[target_automation+2]:has_point_at(line) then
          target_pattern_track.automation[target_automation+2]:remove_point_at(line)
        end
      else
        --dly
        if line_dly <= 1 and line_dly >= 0 then
          line_dly = wte_zoom_env_lower_half(line_dly)
          target_pattern_track.automation[target_automation+2]:add_point_at(line,line_dly) 
        end
      end  
    end
  end --"DLY"----
  
  --addnotifiers
  add_notifiers_and_timers()
end--function
------------------------------------------------------------------------------





--(envelope -> pattern)
-----------------------------------------------------------------------------------------
--this function is called by the timer
--if the flag is set to false it returns early else it writes the automation data to the
--pattern editor
-----------------------------------------------------------------------------------------
function write_envelope_to_pattern()

  --see if the dialog is not visible then exit the tool (test is done within the function) 
  close_gui_on_invisible()
    
  --check flag and return early if false
  if not flag then
    return
  end
  
   --renoise.song()
  local song = renoise.song()
  --return if no parameter selected
  if not song.selected_parameter then
    return
  end
  
  --remove notifiers so nothing fires them within function
  --i.e. automation points notifiers 
  remove_notifiers_and_timers()
    
 
  --selected indexes
  local pattern = song.selected_pattern_index
  local track = song.selected_track_index
  local column = song.selected_note_column_index
  local start_col = column
  local end_col = start_col
  
  --all columns or just selected (loop parameters)
  if flag_all_columns then
    start_col = 1
    end_col = 12
  end
        
  --selected pattern_track + its length
  local selected_pattern_track = song.patterns[pattern].tracks[track]
  local sel_pattern_lines = renoise.song().patterns[pattern].number_of_lines
  
  --return early if not on a sequencer column
  if not song.selected_note_column then
    return
    renoise.app():show_status("Meta From Automation: (Trying to write to fx column)") 
  end

  --loop through automations (stack?) to get the automation number for the "Velocity Device" - Input parameter 
  local target_automation = nil
  for i = 1, #selected_pattern_track.automation do
    if selected_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
      if selected_pattern_track.automation[i].dest_parameter.name == "VEL" then
        target_automation = i
      end
    end
  end

  --if the Input envelope on hydra is empty return early
  if target_automation == nil then
    renoise.app():show_status("Meta From Automation: (Input envelope not set)") 
    return
  end
  
  --REMOVE SUB LINE POINTS IF PRESENT i.e. IF AUTOMATION IS ZOOMED
  -------------------------------------------------------------
  --loop checks if points are present between the lines with the modulus operator
  --These  points have a fraction of line as their time value (1/256) and are present when automation editor is zoomed
  --For example a time of 1.5 means: line 1 with a note column delay of 128.
  local sub_line_points_found = false
  local total_points = #selected_pattern_track.automation[target_automation].points
  local point_counter = total_points
 
  --loop backwards as when we remove points the table shrinks.  if looping forwards the indexes would
  --change as each point was removed
  for i = total_points,1,-1 do
    --if there is a point here
    if selected_pattern_track.automation[target_automation].points[i] then
      local point_time = selected_pattern_track.automation[target_automation].points[point_counter].time
      --decrement counter
      point_counter = point_counter - 1
      if (point_time % 1) ~= 0 then
        sub_line_points_found = true
        selected_pattern_track.automation[target_automation]:remove_point_at(point_time)  
      end
    end
  end
  
  --the automation[] is a table with a length of the number of points present
  --i.e for points at {3,5,6} the table length is only three.  We need a counter to keep track of this 
  local auto_counter = 1

  --loop to copy automation to pattern editor from envelope
  for line = 1,(sel_pattern_lines ) do

    if song.selected_parameter.name == "VEL" then
      local line_vel = 1 --default empty  
      --if there is a point present then set line_vel
      if selected_pattern_track.automation[target_automation]:has_point_at(line) then
        --if the point is at maximum set line_val to 255 this gives full volume "--" empty notation rather than,
        -- equivalent but more messy 80 in the vol column 
        if selected_pattern_track.automation[target_automation].points[auto_counter].value == 1 then
          line_vel = VEL_EMPTY
        else
          line_vel = (selected_pattern_track.automation[target_automation].points[auto_counter].value) * VEL_MAX
        end
        auto_counter = auto_counter + 1
      end
      
   
      
      
      --**** write line_vel to track ***
      --filter note offs and empty notes
      if selected_pattern_track:line(line).note_columns[column].note_string == "OFF" or
         selected_pattern_track:line(line).note_columns[column].note_string == "---" then
        --do nothing in noteoff or empty line
      else 
 
     
     
           
     line_vel =  wtp_zoom_env_lower_half(line_vel)



        --**** write line_vel to track ***
        selected_pattern_track:line(line).note_columns[column].volume_value = line_vel
        
        --if all columns checkbox option selected then copy relevant values across to other columns
        if end_col ~= start_col then
          for col = start_col,end_col do
            --**** write line_vel to track for all other columns ***
            if selected_pattern_track:line(line).note_columns[col].note_string == "OFF" or
              selected_pattern_track:line(line).note_columns[col].note_string == "---" then
              --do nothing in noteoff or empty line
            else 
             
              selected_pattern_track:line(line).note_columns[col].volume_value  = line_vel 
              
            end
          end 
        end
      end 
      
    end
    
    if song.selected_parameter.name == "PAN" then
      local line_pan = 0.5 --default mid line 
      --if there is a point present then set line_vel
      if selected_pattern_track.automation[target_automation+1]:has_point_at(line) then
        --if the point is at 0.5 set line_val to 255 this gives center pan "--" empty notation rather than,
        -- equivalent but more messy 00 in the vol column 
        if selected_pattern_track.automation[target_automation+1].points[auto_counter].value == 0.5 then
          line_pan = PAN_EMPTY
        else
          line_pan = (selected_pattern_track.automation[target_automation+1].points[auto_counter].value) * PAN_MAX
        end
        auto_counter = auto_counter + 1
      end
      
       line_pan =  wtp_zoom_env_lower_half(line_pan)
    --**** write line_vel to track ***
      if selected_pattern_track:line(line).note_columns[column].note_string == "OFF" or
         selected_pattern_track:line(line).note_columns[column].note_string == "---" then
        --do nothing in noteoff or empty line
      else 
        selected_pattern_track:line(line).note_columns[column].panning_value = line_pan
        
        --if all columns checkbox option selected then copy relevant values across to other columns
        if end_col ~= start_col then
          for col = start_col,end_col do
            --**** write line_pan to track for all other columns ***
            if selected_pattern_track:line(line).note_columns[col].note_string == "OFF" or
              selected_pattern_track:line(line).note_columns[col].note_string == "---" then
              --do nothing in noteoff or empty line
            else 
              selected_pattern_track:line(line).note_columns[col].panning_value = line_pan
            end
          end 
        end
      
      end
    end
    
    if song.selected_parameter.name == "DLY" then
      local line_dly = 0 --default mid line 
      --if there is a point present then set line_vel
      if selected_pattern_track.automation[target_automation+2]:has_point_at(line) then
        --if the point is at 0 set line_dly to 255 this gives delay "--" empty notation rather than,
        -- equivalent but more messy 00 in the delay column 
        if selected_pattern_track.automation[target_automation+2].points[auto_counter].value == 0 then
          line_dly = DLY_EMPTY
        else
          line_dly = (selected_pattern_track.automation[target_automation+2].points[auto_counter].value) * DLY_MAX
        end
        auto_counter = auto_counter + 1
      end
      
      line_dly =  wtp_zoom_env_lower_half(line_dly)
      
      --**** write line_del to track ***
      if selected_pattern_track:line(line).note_columns[column].note_string == "OFF" or
         selected_pattern_track:line(line).note_columns[column].note_string == "---" then
        --do nothing in noteoff or empty line
      else 
        selected_pattern_track:line(line).note_columns[column].delay_value = line_dly
          
               
        --if all columns checkbox option selected then copy relevant values across to other columns
        if end_col ~= start_col then
          for col = start_col,end_col do
            --**** write line_pan to track for all other columns ***
            if selected_pattern_track:line(line).note_columns[col].note_string == "OFF" or
              selected_pattern_track:line(line).note_columns[col].note_string == "---" then
            
              --do nothing in noteoff or empty line
            else 
              selected_pattern_track:line(line).note_columns[col].delay_value = line_dly
            end
          end 
        end
      end
    end     
    
 
 
  end --for
 

  --when checkbox enabled on GUI so all columns are updated to the current envelope values
--  if flag_all_columns then
 --   copy_current_envelope_to_all_columns()  ____UPDATE THIS____
 -- end
  
  
  --Now Reset the envelope values that are not associated with a pattern note to zero
  --i.e. the ones for "---" pattern lines
  -----------------------------------------
    --get the values from the pattern for the automation envelope
    --loop to --GET POINTS FROM PATTERN-- editor and write them to the envelope
    ---------------------------------------------------------------------------- 

   for line = sel_pattern_lines, 1,-1 do
      
      if song.selected_parameter.name == "VEL" then   
      --get velocity from pattern
      local line_vel = selected_pattern_track:line(line).note_columns[column].volume_value/VEL_MAX
      
      --255 means empty so remove point from automation
      if line_vel == (VEL_EMPTY/VEL_MAX) then
        line_vel = 1
      end
       --remove points if no note is present
       if selected_pattern_track:line(line).note_columns[column].note_string == "OFF" or
         selected_pattern_track:line(line).note_columns[column].note_string == "---" then 
         if selected_pattern_track.automation[target_automation]:has_point_at(line) then
           selected_pattern_track.automation[target_automation]:remove_point_at(line)
         end
       else 
         --scale value
         line_vel = wte_zoom_env_lower_half(line_vel)
         
         --write the pattern velocity to the envelope value
         selected_pattern_track.automation[target_automation]:add_point_at(line, line_vel)  
       end  
      end
      
      if song.selected_parameter.name == "PAN" then   
      --get velocity from pattern
      local line_pan = selected_pattern_track:line(line).note_columns[column].panning_value/PAN_MAX
      
      --255 means empty so remove set automation to mid point 0.5
      if line_pan == (PAN_EMPTY/PAN_MAX) then
        line_pan = 0.5
      end
       --remove points if no note is present
       if selected_pattern_track:line(line).note_columns[column].note_string == "OFF" or
         selected_pattern_track:line(line).note_columns[column].note_string == "---" then 
         if selected_pattern_track.automation[target_automation+1]:has_point_at(line) then
           selected_pattern_track.automation[target_automation+1]:remove_point_at(line)
         end
       else 
         --scale value
         line_pan = wte_zoom_env_lower_half(line_pan)
         --write the pattern velocity to the envelope value
         selected_pattern_track.automation[target_automation+1]:add_point_at(line, line_pan)  
       end  
      end
      
      if song.selected_parameter.name == "DLY" then   
      --get velocity from pattern
      local line_dly = selected_pattern_track:line(line).note_columns[column].delay_value/DLY_MAX
      
      --255 means empty so remove set automation to mid point 0.5
      if line_dly == (DLY_EMPTY/DLY_MAX) then
        line_dly = 0
      end
       --remove points if no note is present
       if selected_pattern_track:line(line).note_columns[column].note_string == "OFF" or
         selected_pattern_track:line(line).note_columns[column].note_string == "---" then 
         if selected_pattern_track.automation[target_automation+2]:has_point_at(line) then
           selected_pattern_track.automation[target_automation+2]:remove_point_at(line)
         end
       else 
        --scale value
         line_dly = wte_zoom_env_lower_half(line_dly)
         --write the pattern velocity to the envelope value
         selected_pattern_track.automation[target_automation+2]:add_point_at(line, line_dly)  
       end  
      end
    
    end --for line = sel_pattern_lines
 
   --reset flag to false as all done for now to stop the whole function triggering again with timer
  flag = false 
  add_notifiers_and_timers() 
end


-----------------------------------------------------------------------------
--function to remove notifiers and timers (for selected track with notifiers)
-----------------------------------------------------------------------------
function remove_notifiers_and_timers()

  --remove timers
  if renoise.tool():has_timer(check_selected_column_changed) then
    renoise.tool():remove_timer(check_selected_column_changed)
  end

  if renoise.tool():has_timer(write_envelope_to_pattern) then 
    renoise.tool():remove_timer(write_envelope_to_pattern)
  end
 
  --remove notifiers
  --get required info
  ---------------------------------------------
  --renoise.song()
  local song = renoise.song()  
  --selected pattern_track 
  local selected_pattern_track = song.selected_pattern_track
  
  --loop through automations (stack?) to get the automation number for the "Velocity Device" - Input parameter
  local target_automation = nil
  for i = 1, #selected_pattern_track.automation do
    if selected_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
      if selected_pattern_track.automation[i].dest_parameter.name == "VEL" then
        target_automation = i
      end
    end
  end
  
  --if the Input Envelope exists then remove any points notifiers
  if target_automation and selected_pattern_track.automation[1] then  --------------FIX---------- needs proper automation number
    ------------------------------------------------------
    --remove notifier if present
    --vel
    if selected_pattern_track.automation[target_automation].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation].points_observable:remove_notifier(set_flag)
    end
    --pan
    if selected_pattern_track.automation[target_automation+1].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation+1].points_observable:remove_notifier(set_flag)
    end
    --dly
    if selected_pattern_track.automation[target_automation+2].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation+2].points_observable:remove_notifier(set_flag)
    end   
  end 

end
------------


------------------------------------------------------------
--function to add notifiers and timers (for selected track with notifiers)
------------------------------------------------------------
function add_notifiers_and_timers()
 
  --add timers vel
  if not renoise.tool():has_timer(check_selected_column_changed) then
    renoise.tool():add_timer(check_selected_column_changed,100)
  end
  
  if not renoise.tool():has_timer(write_envelope_to_pattern) then 
    renoise.tool():add_timer(write_envelope_to_pattern,300)
  end
  
  --renoise.song()
  local song = renoise.song()  
  --selected pattern_track 
  local selected_pattern_track = song.selected_pattern_track
  
  --loop through automations (stack?) to get the automation number for the "Velocity Device" - Input parameter
  local target_automation = nil
  for i = 1, #selected_pattern_track.automation do
    if selected_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
      if selected_pattern_track.automation[i].dest_parameter.name == "VEL" then
        target_automation = i
      end
    end
  end
   
  --add notifiers
  --if the Input Envelope exists then re-add any points notifiers
  if target_automation and selected_pattern_track.automation[1] then
    --vel
    if not selected_pattern_track.automation[target_automation].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation].points_observable:add_notifier(set_flag)
    end
    --pan
    if not selected_pattern_track.automation[target_automation+1].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation+1].points_observable:add_notifier(set_flag)
    end
    --dly
    if not selected_pattern_track.automation[target_automation+2].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation+2].points_observable:add_notifier(set_flag)
    end
  end

end








---------------------------------------------------------------
--timer checks if the selected column has changed
--if it has then the envelope is updated to show the
--vol/pan/delay data for the newly selected column
---------------------------------------------------------------

function check_selected_column_changed()
  
  local song = renoise.song()
  local trk = song.selected_track_index
  
  --if no selected parameter then return otherwise parmater.name below will 
  --be a nil value
  if not song.selected_parameter then
    return
  end
  
  --if no change return
  if song.selected_note_column_index == g_selected_column and
     song.selected_track_index == g_selected_track        and
     song.selected_pattern_index == g_selected_pattern    and
     song.selected_parameter.name == g_selected_parameter_name then 
    return 
  else
    
    --update to new column and track indexes 
    g_selected_column = renoise.song().selected_note_column_index
    g_selected_track = renoise.song().selected_track_index
    g_selected_pattern = song.selected_pattern_index
    g_selected_parameter_name = song.selected_parameter.name
    
    --get values for notifier handling
    local selected_pattern_track = song.selected_pattern_track
    
    -- loop through devices to make sure automation is set up.  i.e. when changing to a new
    -- pattern the automation envelope will not have been set up.
      
    --if no automation exists then create it.--this can occur on pattern change
    for dev = 2,#song.tracks[trk].devices do
      if song.tracks[trk].devices[dev].display_name == "Vel Pan Dly" then
        --create automations
        if not selected_pattern_track:find_automation(song.tracks[trk].devices[dev].parameters[1]) then 
          selected_pattern_track:create_automation(song.tracks[trk].devices[dev].parameters[1])--vel
        end
        if not selected_pattern_track:find_automation(song.tracks[trk].devices[dev].parameters[2]) then 
          selected_pattern_track:create_automation(song.tracks[trk].devices[dev].parameters[2])--pan
        end
        if not selected_pattern_track:find_automation(song.tracks[trk].devices[dev].parameters[3]) then 
          selected_pattern_track:create_automation(song.tracks[trk].devices[dev].parameters[3])--dly          
        end
      end
    end

    --loop through automations (stack?) to get the automation number for the "Velocity Device" - Input parameter
    local target_automation = nil
    for i = 1, #selected_pattern_track.automation do
      if selected_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
        if selected_pattern_track.automation[i].dest_parameter.name == "VEL" then
          target_automation = i
        end
      end
    end
    
    --when a new track is added while the tool is running then there will be no device so add one
    if not target_automation then
      create_automation_envelopes_all_tracks()
      --re-check and get target automation
      for i = 1, #selected_pattern_track.automation do
        if selected_pattern_track.automation[i].dest_device.display_name == "Vel Pan Dly" then
          if selected_pattern_track.automation[i].dest_parameter.name == "VEL" then
            target_automation = i
          end
        end
      end
    end
    
    --return if there is no automation envelope set up as new track could be as send etc.
    if target_automation == nil then
      return
    end
    
    --set automation playmode to points for vel pan dly envelopes
    selected_pattern_track.automation[target_automation].playmode = renoise.PatternTrackAutomation.PLAYMODE_POINTS --vel
    selected_pattern_track.automation[target_automation+1].playmode = renoise.PatternTrackAutomation.PLAYMODE_POINTS --pan
    selected_pattern_track.automation[target_automation+2].playmode = renoise.PatternTrackAutomation.PLAYMODE_POINTS --dly  
    
    --------------------
    --check if selected envelope is vol,pan,dly, if not return (this is also done in update_selected_envelope_from_pattern )
    local p_name
    --check if parameter selected if not return
    if song.selected_parameter then
      p_name = song.selected_parameter.name
    else
      renoise.app():show_status("`Vel Pan Dly` Tool: (No Parameter Selected, please select one)") 
      return
    end
    
    --return if not at least one of the following envelope names is the selected parameter(==envelope)
    if not ((p_name == "VEL") or (p_name == "PAN") or (p_name == "DLY")) then
      return
    end
    -----------------------
    
    --if we are on pan then set target automation accordingly +1
    if p_name == "PAN" then
      target_automation = target_automation + 1
    elseif  p_name == "DLY" then  
      target_automation = target_automation + 2
    end
    

    
    --disable automation points notifier so it doesn`t fire when points change here
    if selected_pattern_track.automation[target_automation].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation].points_observable:remove_notifier(set_flag)
    end
    
    --update the envelope
    update_selected_envelope_from_pattern()
    
    --reinstate automation points notifier
    if not selected_pattern_track.automation[target_automation].points_observable:has_notifier(set_flag) then
      selected_pattern_track.automation[target_automation].points_observable:add_notifier(set_flag)
    end
  end 
end

-----------------------------------------------------------------------
--Alternative undo
--releases all notifiers,does: undo() and then re-initialises notifiers
-----------------------------------------------------------------------
function alt_undo(undo)
  --remove notifiers
  remove_notifiers_and_timers()
  remove_notifiers_and_timers()
  --undo/ redo
  if undo then
    renoise.song():undo()
  else
    renoise.song():redo()
  end
  --add notifiers
   add_notifiers_and_timers()
   add_notifiers_and_timers()
end

----------------------------------------------------------------
--flag function sets flag to true so pattern will be updated in 
--write_envelope_to_pattern() function
----------------------------------------------------------------
function set_flag(bool)
  flag = true
end

-------------------------------------------------
--function to check for presence of dummy devices
-------------------------------------------------
function check_for_present_dummy_devices()
  local song = renoise.song()
  --loop tracks
  for trk = 1, #song.tracks do
    --loop to check if "Vel Pan Dly" DSP has already been added on each track
    local found_device = false
    for i = 2,#song.tracks[trk].devices do
      if song.tracks[trk].devices[i].display_name == "Vel Pan Dly" then
        return true
      end
    end
  end
  return false
end

----------------------------------------------------------------------
--Removes all dummy devices and envelopes, releases timers
----------------------------------------------------------------------
function remove_dummy_devices()

  --reset global flag all columns to false in case needed
  flag_all_columns = false
  flag = false
   
  local song = renoise.song()
  
  --remove timers that the tool is firing
  --1 remove timers 
  --(pattern -> envelope)
  if renoise.tool():has_timer(check_selected_column_changed) then
    renoise.tool():remove_timer(check_selected_column_changed)
  end
  --(envelope -> pattern)
  if renoise.tool():has_timer(write_envelope_to_pattern) then 
   renoise.tool():remove_timer(write_envelope_to_pattern) 
  end
   
    
  --delete devices, when the device is deleted renoise takes care of releasing it`s notifiers aswell.
  for trk = 1,#song.tracks do
    for dev = 1,#song.tracks[trk].devices do
      if song.tracks[trk].devices[dev].display_name == "Vel Pan Dly" then
        song.tracks[trk]:delete_device_at(dev)  
        break
      end
    end
  end
end

------------------------------------------------------------------------------
--called by slider to set all values in the current
--column
----------------------------------------------------------------------
function set_all_values_in_column(val)

  --renoise vals
  local song = renoise.song()
  local p_track = song.selected_pattern_track
  local start_col = song.selected_note_column_index 
  local end_col = start_col
  
  --return early if not a note column
  if not song.selected_note_column then
    return
  end
  
  --check if selected envelope is vol,pan,dly, if not return (this is also done in update_selected_envelope_from_pattern )
  local p_name
  --check if parameter selected if not return
  if song.selected_parameter then
    p_name = song.selected_parameter.name
  else
    renoise.app():show_status("`Vel Pan Dly` Tool: (No Parameter Selected, please select one)") 
    vb.views["val text"].text = "Select VEL, PAN or DLY ENVELOPE"
    return false
  end
  
  --return if not at least one of the following envelope names is the selected parameter(==envelope)
  if not ((p_name == "VEL") or (p_name == "PAN") or (p_name == "DLY")) then
    --update status on GUI
    vb.views["val text"].text = "Select VEL, PAN or DLY ENVELOPE"
    return false
  end
  
  --remove notifiers
  remove_notifiers_and_timers()
  
  --all columns or just selected
  if flag_all_columns then
    start_col = 1
    end_col = 12
  end
  
  if p_name == "VEL" then
    --set the sliders maximum value to 128
    vb.views["slider"].max = VEL_MAX
    --if at 128 max value make pattern editor the default clear "--" 255
    if val == VEL_MAX then
      val = VEL_EMPTY
    end
  end
  
  if p_name == "PAN" then
    --set the sliders maximum value to 128
    vb.views["slider"].max = PAN_MAX
    --if at centre value make pattern editor the default clear
     if val == PAN_CENTRE then
      val = PAN_EMPTY --255
    end
  end
  
  if p_name == "DLY" then
    --delay has a larger range of 255
    vb.views["slider"].max = DLY_MAX
  end
  
  -- loop through pattern
  for line = 1,#p_track.lines do 
    for col = start_col,end_col do 
      --update pattern-- only lines with notes on them
      if p_track:line(line).note_columns[col].note_string == "OFF" or
         p_track:line(line).note_columns[col].note_string == "---" then  
        --do nothing in noteoff or empty line
      else
        if p_name == "VEL" then
          p_track:line(line).note_columns[col].volume_value = val
        elseif p_name == "PAN" then
          p_track:line(line).note_columns[col].panning_value = val
        elseif  p_name == "DLY" then 
          p_track:line(line).note_columns[col].delay_value = val 
        end
      end
    end
  end
  
  --update envelope
  update_selected_envelope_from_pattern()
  --re--add notifiers
 -- add_notifiers_and_timers()
  --reset pan value if it is centre, so that it will display properly on the GUI
  if val == PAN_EMPTY then
     val = PAN_CENTRE 
  end
  --update GUI string with formatted value (+ space to centre)
  --make sure text will be added centrally, other messages may reset this
  vb.views["val text"].align = "center"
  vb.views["val text"].text = num_to_hex_string(val).." ("..string.format("%u",tostring(val))..")" 
  
   --re-add notifiers
  add_notifiers_and_timers()
end





--END of Functions_non_spec.lua
----------------------------------------------------------------------
